---
title: LLRB tree invariants
nav_order: 4
---

Converting from the left-leaning BST back to a 2-3 tree is somewhat tricky since it's hard to tell which nodes represent 3-nodes rather than 2-nodes. This is also hard to do in code too, so left-leaning red-black trees introduce a little bit of extra information to the "glue" edge.

Left-leaning red-black (LLRB) trees color "glue" edges (representing 3-nodes) red. Red edges connect separated 3-nodes and help us to maintain correspondence with 2-3 trees but otherwise don't have any special functionality. In code, the red edge is represented in code as an extra `color` field stored in the left-leaning **node**. Remember that edges in a binary search tree are just references, so they don't have fields. Only nodes are objects, so any extra information such as red/black coloring needs to be stored in a node.

1-1 correspondence
: Every 2-3 tree has a unique left-leaning red-black tree associated with it.

LLRB tree invariants follow entirely from 1-1 correspondence with 2-3 trees.

- Red edges lean left---all 3-nodes are represented as two nodes connected with left-leaning "glue" edges.
- No node has two red edges connected to it---2-3 trees only allow up to 3-nodes with 2 keys and 3 children.
- Every root-to-null path has the same number of black edges---all 2-3 tree leaf nodes are the same depth from the root.

{% capture question %}
Which of the following are valid LLRB trees?

- [ ] ![Tree 1](https://docs.google.com/drawings/d/e/2PACX-1vSVi0hj6f2LP_KrlNdurZGtW66oFO6_rPl3-EH5HLa0vlGUk90IAAaU8lgPNCNRhvumGKi_j9cO0pyZ/pub?w=288&h=288)
- [ ] ![Tree 2](https://docs.google.com/drawings/d/e/2PACX-1vSJBubOE3bKX-s9I1fAISn5ldXsNCOmpFfnZZh5CRuI0WGUqjdKi9dh5AB-wTJG8dVi9gutU0Ywu5Ms/pub?w=288&h=288)
- [ ] ![Tree 3](https://docs.google.com/drawings/d/e/2PACX-1vTVu7HCOJEOC2ZBSIy14HzxMRlgeEot17LRsYKclnt6Wt7FT9HX-ox95L5OLKmKVZXJnBxemtUlVpcQ/pub?w=288&h=288)
- [ ] ![Tree 4](https://docs.google.com/drawings/d/e/2PACX-1vRbi_LC04lX8B26FFPU1XUEFtjH72-97blS-Ba5XEgpR5KXRC5R9KLFZIBBcyD36Q4XtUUvWVwUN_WD/pub?w=288&h=288)
{% endcapture %}
{% capture explanation %}
Either apply the invariants or convert back to a 2-3 tree.
{% endcapture %}
{% include question.html question=question explanation=explanation number=1 %}

{% capture question %}
What is the height of the 2-3 tree corresponding to this LLRB tree?

![Tree 4](https://docs.google.com/drawings/d/e/2PACX-1vRbi_LC04lX8B26FFPU1XUEFtjH72-97blS-Ba5XEgpR5KXRC5R9KLFZIBBcyD36Q4XtUUvWVwUN_WD/pub?w=288&h=288)
{% endcapture %}
{% include question.html question=question number=2 %}
