---
title: Merge sort
nav_order: 5
---

Merge sort is a recursive sorting algorithm that we'll use as a case study of recurrence relations. It relies on the **merge** operation, which takes two **sorted** arrays and returns a sorted result containing all of the items in both of the given arrays.

Merge sort
: 1. If the array is of size 1, return.
  2. Recursively merge sort the left half.
  3. Recursively merge sort the right half.
  4. Merge the two sorted halves.

Recurrence relation for the runtime of merge sort
: *T*(*N*) = *T*(*N*/2) + *T*(*N*/2) + *c*<sub>1</sub>*N* + *c*<sub>0</sub> for *N* > 1
: *T*(1) = *d*

Each call to merge sort makes two recursive calls to subproblems of half the size each and then spends linear time merging the sorted halves. Unrolling this recurrence is a bit trickier since there are two recursive branches. Another approach is to draw a **recurrence diagram** that represents the recursive work and non-recursive work in the recurrence relation!

![Merge Sort](https://docs.google.com/drawings/d/e/2PACX-1vQLIGT0uu9k3Ouz-QGHsotJg6upr8r8Jn5JO9z2UQtv32aCrEimtcxHvmq-wvEGbML7NRIZIjb_e6oZ/pub?w=1440&h=810)

The top layer takes about 64 units of time merging two sorted halves of 32 items each. The second layer also takes about 64 units to time merging four sorted halves of 16 items each. By identifying the pattern in the recurrence diagram, we can see that each level will take about 64 units of time.

Level
: All of the nodes at the same depth (number of edges) from the overall root.

Since the entire runtime of merge sort is represented by this diagram, we can find the total time spent by multiplying the number of levels in the tree by the time spent on each level. Each level divides the input size in half until the base case of 1 item is reached so there are log<sub>2</sub>*N* levels. Each level takes about N time to merge many small arrays. Overall, the order of growth of the runtime for merge sort is in N log<sub>2</sub>*N*.
