---
title: Recurrence diagrams
nav_order: 7
---

Merge sort is an example of a **work-per-level recurrence**. We found that each level in the recurrence diagram performed roughly the same amount of (non-recursive) work. The sum of all of the levels resulted in the total amount of time merge sort takes to run, including all recursive work (each node in the diagram) and all of the non-recursive work (the merge operation). Two big assumptions enabled us to use multiplication as a shortcut to simplify the final result.

1. What is the (non-recursive) work done by each node on level *i*? How many total nodes are on each level *i*? **Assumption 1**: Each node on level *i* performs the same amount of work, so we can multiply the work per node by the number of nodes on that level.
1. What is the total (non-recursive) work done on level *i*? How many levels are in the tree? **Assumption 2**: Each level performs the same amount of work, so we can multiply the work per level by the number of levels to get the total runtime of the entire recurrence.

Even when these assumptions don't hold, we can still go back to writing out the sum of the terms instead of using multiplication as a shortcut. **Work-per-node recurrences** represent a category of recurrences that don't meet assumption 2. Consider `f3`.

```java
static int f3(int n) {
    if (n <= 1)
        return 1;
    return f3(n - 1) + f3(n - 1);
}
```

{% capture question %}
Give a recurrence relation for the runtime of `f3`. Try drawing out the first couple levels of the recurrence diagram!
{% endcapture %}
{% capture explanation %}
*T*(*N*) = 2*T*(*N* − 1) + *c*
: 2*T*(*N* − 1): Two recursive calls to f3. Alternatively, *T*(*N* − 1) + *T*(*N* − 1).
: *c*: non-recursive work like comparing, plus, and return.

*T*(1) = *d*
: *d*: compare and return 1.

![f3](https://docs.google.com/drawings/d/e/2PACX-1vRP1iJS-5hvT-ErJQ4-PH4esff-x9RaZyljhLpKewAhDiOj7IDWrxm5lUyjEoRF4305Vy6Akh_Qh9oq/pub?w=960&h=540)

Optionally, we can solve the recurrence relation directly at this point and find the total runtime for `f3` with respect to *N*. In the questions below, we’ll walkthrough this logic from a different direction by relying on the simplifying assumptions above. You might prefer one approach over the other. Here, we’ll simplify *c* and *d* and just treat them as 1 since our asymptotic analysis doesn’t care about the difference between constants.

1. *T*(*N*) = 2*T*(*N* − 1) + 1
1. *T*(*N*) = 2\[2*T*(*N* − 2) + 1\] + 1
1. *T*(*N*) = 2\[2\[2*T*(*N* − 3) + 1\] + 1\] + 1
1. *T*(*N*) = 2<sup>*N* − 1</sup> + 2<sup>*N* − 2</sup> + 2<sup>*N* − 3</sup> + ⋯ + 2<sup>0</sup>
1. *T*(*N*) = 1 + 2 + 4 + ⋯ + 2<sup>*N* − 1</sup>
1. *T*(*N*) = 2<sup>*N*</sup> − 1
1. *T*(*N*) ∈ *Θ*(2<sup>*N*</sup>)
{% endcapture %}
{% include question.html question=question explanation=explanation number=1 %}

{% capture question %}
What is the (non-recursive) work done by each node on level *i*?

- [ ] Constant
- [ ] Logarithmic with respect to the size of the current subproblem.
- [ ] Linear with respect to the size of the current subproblem.
{% endcapture %}
{% capture explanation %}
If we use our explanation from the recurrence relation, the non-recursive work for both *T*(*N*) and *T*(1) are represented by the constants *c* and *d*, respectively.
{% endcapture %}
{% include question.html question=question explanation=explanation number=2 %}

{% capture question %}
How many total nodes are on each level *i*?

Assume the overall root node is at level 0, its immediate children at level 1, and so forth.

- [ ] 1
- [ ] *i*
- [ ] 2*i*
- [ ] 2<sup>*i*</sup>
{% endcapture %}
{% capture explanation %}
Each call to `f3` creates two more calls to `f3`! Therefore, on level *i*, there will be 2<sup>*i*</sup> nodes.
{% endcapture %}
{% include question.html question=question explanation=explanation number=3 %}

{% capture question %}
Although assumption 1 is true and we can multiply the answers to the above two questions to get the work-per-level, we'll see that the work-per-level is not all the same.

Give a closed-form expression for *T*(*N*) as a sum of all of the levels in the recurrence diagram. For simplicity, replace the constants *c* and *d* with 1.

Then, apply a mathematical identity to find a simple big-theta runtime bound.
{% endcapture %}
{% capture explanation %}
Although assumption 2 doesn't hold, we can still use the sum of successive powers of 2 to solve the problem. This is a work-per-node recurrence analysis because what we've ultimately done is counted the number of nodes in the recurrence diagram and multiplied it by the (non-recursive) work done per node. There are 2<sup>*N*</sup> − 1 nodes in the diagram and each node performs a constant amount of work.
{% endcapture %}
{% include question.html question=question explanation=explanation number=4 %}
